#include <A4Stuff.h>
#include <MixedMode.h>
#include <QDOffscreen.h>
#include <TextUtils.h>


#define	kFlowerHeight		43
#define	kFlowerBorder		1
#define	kMaxFrames			10
#define	kBlankTime			7	// Time between create new flower checks
#define	kFlowerTime			2	// Time between flower frames
#define	kUp					1
#define	kDown				0


typedef struct
{
	short			color;
	short			curFrame;
	short			going;	// Getting bigger or smaller? (kUp or kDown)

	long			lastChange;	// Last time we drew a frame

	Rect			position;
} Flower, *FlowerPtr;

typedef struct Stem
{
	FlowerPtr		flower;
	Stem			*next;
} *StemPtr;


// Globals
Boolean			gInitialized;
GWorldPtr		gFlowerWorld;
long			gLastTime;		// Last time we checked to see if we should make a flower
Point			gMousePos;
Rect			gDirtyRect;
StemPtr			gRoot;


// Prototypes
void	FlowerAppLoop(void);

void	LoadFlowers(void);
void	CreateNewFlower(void);
void	DestroyFlower(void);
void	UpdateFlowers(void);
StemPtr	GetLastFlower(void);
void	CreateDirtyRect(Rect *dirtyRect);
void	DrawFrame(StemPtr	temp);

GWorldPtr	CreateOffscreen(Rect GWorldRect);
void		SetupOffscreen(GWorldPtr GWorld);
void 		CopyOver(GWorldPtr GWorld, WindowPtr myWindow, Rect WindowRect, Rect GWorldRect);
void 		CopyMaskOver(GWorldPtr GWorld, WindowPtr myWindow, Rect WindowRect, Rect GWorldRect);
void		CreateDirtyRect(void);

////////////////

enum {
	getNextEventProcInfo = (kPascalStackBased | RESULT_SIZE(sizeof(Boolean)) | 
					STACK_ROUTINE_PARAMETER(1, SIZE_CODE(sizeof(short))) |
					STACK_ROUTINE_PARAMETER(2, SIZE_CODE(sizeof(EventRecord *)))
	)
};

RoutineDescriptor *gGetNextEventAddr;
pascal Boolean MyGetNextEvent(short eventMask, EventRecord *theEvent);
PicHandle		gPicHandle = NULL;

void main(void);
RoutineDescriptor MainRD = BUILD_ROUTINE_DESCRIPTOR(kCStackBased, main);

void main(void)
{
	UniversalProcPtr	theProcPtr;
	THz 				theZone;
	
	EnterCodeResource();
	
	
	DetachResource(GetResource('INIT',2));
	gPicHandle = GetPicture(128);
	HNoPurge((Handle) gPicHandle);
	DetachResource((Handle) gPicHandle);
	
	theZone = GetZone();
	SetZone(SystemZone());


	gGetNextEventAddr = NGetTrapAddress(_GetNextEvent, ToolTrap);
	theProcPtr = NewRoutineDescriptor( (ProcPtr)MyGetNextEvent,getNextEventProcInfo, GetCurrentArchitecture() );
	NSetTrapAddress(theProcPtr, _GetNextEvent, ToolTrap);


	SetZone(theZone);
	
	gInitialized = false;
	gLastTime = 0;
	gRoot = 0;
	
	ExitCodeResource();
}

pascal Boolean MyGetNextEvent(short eventMask, EventRecord *theEvent)
{
	Boolean 	result;
	Point		old;
	
	EnterCodeResource();
		
	result = CallUniversalProc((UniversalProcPtr)gGetNextEventAddr, getNextEventProcInfo, eventMask, theEvent);
	
	// Don't do some quickdraw stuff during startup so we do it the first time our patch is called.
	if(!gInitialized)
	{
		LoadFlowers();
		GetMouse(&gMousePos);
		LocalToGlobal(&gMousePos);
				
		gInitialized = true;
		
//		DebugStr("\pInitialized.");
	}
	
	if(gLastTime + kBlankTime < TickCount())	// Time to see if we should create a new flower?
	{
		old.h = gMousePos.h;
		old.v = gMousePos.v;
		GetMouse(&gMousePos);
		LocalToGlobal(&gMousePos);	

		if(old.h != gMousePos.h || old.v != gMousePos.v)	// if the mouse has moved
			CreateNewFlower();
		
		gLastTime = TickCount();
	}
	UpdateFlowers();
	
	ExitCodeResource();

	return result;
}

#pragma mark -

void	LoadFlowers(void)
{
	Rect		picRect;
	
	
	EnterCodeResource();
	
	picRect = (*gPicHandle)->picFrame;
	OffsetRect(&picRect, - picRect.left, - picRect.top);
	
	gFlowerWorld = CreateOffscreen(picRect);
	DrawPicture(gPicHandle, &picRect);
	
	ExitCodeResource();
}

void	CreateNewFlower(void)
{
	StemPtr	newFlower, last = GetLastFlower();
//	DebugStr("\pCreateNewFLower ");
	EnterCodeResource();
		
	newFlower = (StemPtr)NewPtr(sizeof(Stem));
	newFlower->flower = (FlowerPtr)NewPtr(sizeof(Flower));
	
	if(last)
		last->next = newFlower;
	else
		gRoot = newFlower;
	
	newFlower->flower->color = Random() % 5;
	newFlower->flower->curFrame = 0;
	newFlower->flower->lastChange = 0;
	newFlower->flower->going = kUp;

	SetRect(&newFlower->flower->position, 0, 0, kFlowerHeight, kFlowerHeight);
	OffsetRect(&newFlower->flower->position, gMousePos.h - 20, gMousePos.v - 20);
	
	
	newFlower->next = 0;
	
	ExitCodeResource();
}

void	DestroyFlower(void)
{
	EnterCodeResource();
	
	StemPtr	temp = gRoot;
	
	gRoot = gRoot->next;
	
	DisposePtr((Ptr)temp->flower);
	DisposePtr((Ptr)temp);
	
	ExitCodeResource();
}

void	UpdateFlowers(void)
{
	StemPtr		temp;
	Boolean		disposed = false;
	
	EnterCodeResource();	
	
	for(temp = gRoot;temp;temp = temp->next)
	{
		if(temp->flower->lastChange + kFlowerTime < TickCount())	// if it is time to update the frame of this flower
		{
			if(temp->flower->going == kUp)
			{
				if(temp->flower->curFrame != kMaxFrames)
					temp->flower->curFrame++;
				else	// it is getting bigger and on the biggest frame
				{
					temp->flower->going = kDown;
					temp->flower->curFrame--;
				}
			}
			else
			{
				if(temp->flower->curFrame != 0)
					temp->flower->curFrame--;
				else		// if it is getting smaller and is at the last frame
				{
					DestroyFlower();
					disposed = true;
				}
			}
			
			if(!disposed)
			{
				DrawFrame(temp);
				temp->flower->lastChange = TickCount();
			}
		}
	}
	
	ExitCodeResource();
}

StemPtr	GetLastFlower(void)
{
	StemPtr	temp, last;
	
	EnterCodeResource();
	
	if(!gRoot)
		return 0;

	for(temp = gRoot;temp;temp = temp->next)
		last = temp;
	
	return last;
	
	ExitCodeResource();
}

void	DrawFrame(StemPtr	temp)
{
	Rect	srcRect;
	GrafPtr	screen;
	RgnHandle	myRgn;
	GrafPtr	savePort;
	
	EnterCodeResource();
	
	CreateDirtyRect();
	myRgn = NewRgn();
	RectRgn(myRgn, &gDirtyRect);
	
	GetCWMgrPort((CGrafPtr *) &screen);
	GetPort(&savePort);
	SetPort(screen);
	
	SetRect(&srcRect, temp->flower->curFrame * kFlowerHeight + temp->flower->curFrame + 1,
					temp->flower->color * kFlowerHeight + temp->flower->color + 1,
					temp->flower->curFrame * kFlowerHeight + temp->flower->curFrame + 1 + kFlowerHeight,
					temp->flower->color * kFlowerHeight + temp->flower->color + 1 + kFlowerHeight);
	
//	CopyOver((CGrafPtr)temp->flower->save, (GrafPtr)temp->flower->buffer, temp->flower->save->portRect, temp->flower->buffer->portRect);
//	CalcVisBehind(FrontWindow(), myRgn);
//	PaintBehind(FrontWindow(), myRgn);
//	DebugStr("\pJust about to copymaskover to screen.");
	
	CopyMaskOver((CGrafPtr)gFlowerWorld, screen, temp->flower->position, srcRect);
//	CopyOver((CGrafPtr)temp->flower->buffer, screen, temp->flower->position, temp->flower->buffer->portRect);
	
	SetPort(savePort);
	DisposeRgn(myRgn);
	
	ExitCodeResource();
}

GWorldPtr CreateOffscreen(Rect GWorldRect)
{
	QDErr		myQDErr;
	GWorldPtr 	worldForMe;
	Str255		error;
		
	myQDErr = NewGWorld( &worldForMe, 16, &GWorldRect, NULL, NULL, 0L );
	if(myQDErr != noErr)
		DebugStr("\pOh shit! Take a break and play Bolo.");
	
	NumToString((long)myQDErr, error);
	//DebugStr(error);
		
	SetupOffscreen(worldForMe);
	
	return worldForMe;
}

void	SetupOffscreen(GWorldPtr GWorld)
{
	LockPixels( GetGWorldPixMap( GWorld ) );
	SetPort((GrafPtr)GWorld);
	EraseRect(&GWorld->portRect);
}

void CopyOver(GWorldPtr GWorld, WindowPtr myWindow, Rect WindowRect, Rect GWorldRect)
{
		CopyBits(
		&(((GrafPtr)GWorld)->portBits), 	//bitmap of gworld
		&(((GrafPtr)myWindow)->portBits),		//bitmap of window
		&(GWorldRect),						//portRect of gworld
		&(WindowRect),						//&(dialog->portRect),				
		0, NULL );
}

void CopyMaskOver(GWorldPtr GWorld, WindowPtr myWindow, Rect WindowRect, Rect GWorldRect)
{
		CopyMask(
		&(((GrafPtr)GWorld)->portBits), 	//bitmap of gworld
		&(((GrafPtr)GWorld)->portBits),		//mask bitmap
		&(((GrafPtr)myWindow)->portBits),		//bitmap of window
		&(GWorldRect),						//portRect of gworld
		&(GWorldRect),					// mask rect
		&(WindowRect)						//&(dialog->portRect),				
		);
}

void	CreateDirtyRect(void)
{
	StemPtr		temp;
	
	EnterCodeResource();
	
	gDirtyRect.left = 900;
	gDirtyRect.right = -1;
	gDirtyRect.top = 900;
	gDirtyRect.bottom = -1;
	for(temp = gRoot;temp;temp = temp->next)
	{	// The garbled mess below just sets the bounds of the dirty rect.
		if(temp->flower->position.left < gDirtyRect.left)
			gDirtyRect.left = temp->flower->position.left;
		if(temp->flower->position.right > gDirtyRect.right)
			gDirtyRect.right = temp->flower->position.right;
		if(temp->flower->position.top < gDirtyRect.top)
			gDirtyRect.top = temp->flower->position.top;
		if(temp->flower->position.bottom > gDirtyRect.bottom)
			gDirtyRect.bottom = temp->flower->position.bottom;
	}
	ExitCodeResource();
}